#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define HTML_HEADER								\
"<!-- Part of the Quick ISO C Reference Manual -->\n"				\
"<html>\n"									\
"<head><title>Online ISO C Reference Documentation</title></head>\n"		\
"<body background=\"../images/bg.gif\">\n"		       			\
"<center>\n"									\
"<img src=\"../images/refdoc.gif\">\n"						\
"<pre>\n"									\
"This documentation is copyright &copy; 1998-2001 Sandro Sigala "		\
"&lt;<a href=\"mailto:sandro@sigala.it\">sandro@sigala.it</a>&gt;.\n"		\
"All rights reserved.\n\n"							\
"Released under the <a href=\"../COPYING\">GNU General Public License</a>.\n"	\
"</pre>\n"									\
"</center>\n"									\
"<hr>\n"

#define MAXLINELEN 1024
typedef enum { false, true } bool;

/*-------------------------------------------------------------------------
 * Miscellaneous functions.
 *-------------------------------------------------------------------------*/

char ungetline_text[MAXLINELEN];
bool ungetline_valid = false;
int lineno = 0;

char *getline(char *buf)
{
	if (ungetline_valid) {
		ungetline_valid = false;
		strcpy(buf, ungetline_text);
		return buf;
	}

	if (fgets(buf, MAXLINELEN, stdin) == NULL)
		return NULL;

	++lineno;
	buf[strlen(buf)-1] = '\0';	/* Strip newline. */
	return buf;
}

void ungetline(char *s)
{
	strcpy(ungetline_text, s);
	ungetline_valid = true;
}

char *filter(char *s)
{
	static char buf[MAXLINELEN], *p;

	for (p = buf; *s != '\0'; ++s)
		if (*s == '<') {
			*p++ = '&';
			*p++ = 'l';
			*p++ = 't';
			*p++ = ';';
		} else if (*s == '&') {
			*p++ = '&';
			*p++ = 'a';
			*p++ = 'm';
			*p++ = 'p';
			*p++ = ';';
		} else
			*p++ = *s;
	*p = '\0';

	return buf;
}

FILE *out_file = NULL;
FILE *ex_file = NULL;
FILE *sym_file = NULL;

void filter_block(char *buf, bool ignore_at);

void filter_alltt(char *buf, bool open)
{
	if (!open)
		fprintf(out_file, "<pre>\n");
	filter_block(buf, false);
	fprintf(out_file, "</pre>\n");
}

void filter_verbatim(char *buf, bool open)
{
	if (!open)
		fprintf(out_file, "<pre>\n");
	while (getline(buf) != NULL) {
		if (buf[0] == '@') {
			ungetline(buf);
			break;
		}
		fprintf(out_file, "%s\n", filter(buf));
		if (ex_file != NULL)
			fprintf(ex_file, "%s\n", buf);
	}
	fprintf(out_file, "</pre>\n");
}

void filter_block(char *buf, bool ignore_at)
{
	char *p;

	while (getline(buf) != NULL) {
		if (buf[0] == '@')
			switch (buf[1]) {
			case 'V':
				filter_alltt(buf, 0);
				continue;
			case '@':
				if (!ignore_at)
					return;
				continue;
			case 'L':
				continue;
			case 'H':
				fprintf(out_file, "%s\n", &buf[3]);
				continue;
			case 't':
			case 'i':
			case 'b':
				break;
			default:
				ungetline(buf);
				return;
			}
		if (buf[0] == '\0')
			fprintf(out_file, "<br>");
		for (p = buf; *p != '\0'; ++p)
			if (*p == '@') {
				int t = *++p;
				switch (t) {
				case 't': fprintf(out_file, "<code>"); break;
				case 'i': fprintf(out_file, "<i>"); break;
				case 'b': fprintf(out_file, "<b>"); break;
				default: fprintf(stderr, "%d: Invalid type `%c'\n", lineno, t);
				}
				++p;
				while (*++p != '}')
					if (*p == '<')
						fprintf(out_file, "&lt;");
					else if (*p == '&')
						fprintf(out_file, "&amp;");
					else
						fputc(*p, out_file);
				switch (t) {
				case 't':
					fprintf(out_file, "</code>");
					break;
				case 'i':
					fprintf(out_file, "</i>");
					break;
				case 'b':
					fprintf(out_file, "</b>");
					break;
				}
			} else if (*p == '<') {
				fprintf(out_file, "&lt;");
				if (ex_file != NULL)
					fputc('<', ex_file);
			} else if (*p == '&') {
				fprintf(out_file, "&amp;");
				if (ex_file != NULL)
					fputc('&', ex_file);
			} else {
				fputc(*p, out_file);
				if (ex_file != NULL)
					fputc(*p, ex_file);
			}
		fputc('\n', out_file);
		if (ex_file != NULL)
			fputc('\n', ex_file);
	}
}

void filter_headers(char *buf)
{
	if (*buf != '\0') {
		char *s;
		s = strtok(buf, " ");
		while (s != NULL) {
			fprintf(out_file, "#include &lt;%s>\n", s);
			if (ex_file != NULL)
				fprintf(ex_file, "#include <%s>\n", s);
			s = strtok(NULL, " ");
		}
		fputc('\n', out_file);
		if (ex_file != NULL)
			fputc('\n', ex_file);
	}
}

/*-------------------------------------------------------------------------
 * Parsing functions.
 *-------------------------------------------------------------------------*/

char header_name[MAXLINELEN];

void do_xref(char *buf)
{
/*	printf("\\Xref{%s}{%s}{sym:%s}\n",
	       header_name, buf, buf);*/
}

void do_prototype(char *buf)
{
	fprintf(out_file, "<h2>Prototype</h2>\n"
	       "<pre>\n");
	filter_headers(buf);
	filter_alltt(buf, true);
}

void do_description(char *buf)
{
	fprintf(out_file, "<h2>Description</h2>\n");
	filter_block(buf, true);
}

static int source_idx = 0;

void do_implementation(char *buf)
{
	char filename[24];
	sprintf(filename, "source_%d.c", ++source_idx);
	ex_file = fopen(filename, "w");
	fprintf(out_file,
		"<h2>Implementation</h2>\n"
		"<a href=\"%s\"><img src=\"../images/view.gif\" alt=\"View source\" border=\"0\" height=\"32\" width=\"32\"></a>\n"
		"<br>\n"
		"<pre>\n", filename);
	filter_headers(&buf[2]);
	filter_verbatim(buf, true);
	fclose(ex_file);
	ex_file = NULL;
}

void do_example(char *buf)
{
	char filename[24];
	sprintf(filename, "source_%d.c", ++source_idx);
	ex_file = fopen(filename, "w");
	fprintf(out_file,
		"<h2>Example</h2>\n"
		"<a href=\"%s\"><img src=\"../images/view.gif\" alt=\"View source\" border=\"0\" height=\"32\" width=\"32\"></a>\n"
		"<br>\n"
		"<pre>\n", filename);
	filter_headers(&buf[2]);
	filter_verbatim(buf, true);
	fclose(ex_file);
	ex_file = NULL;
}

static bool done_1990=false, done_1999=false;

void do_iso1990_ref(char *buf)
{
	if (done_1999)
		fprintf(out_file, ", ISO C 9899:1990 <b>%s</b>", buf);
	else
		fprintf(out_file, "<h2>References</h2>\nISO C 9899:1990 <b>%s</b>", buf);
	done_1990 = true;
}

void do_iso1999_ref(char *buf)
{
	if (done_1990)
		fprintf(out_file, ", ISO C 9899:1999 <b>%s</b>", buf);
	else
		fprintf(out_file, "<h2>References</h2>\nISO C 9899:1999 <b>%s</b>", buf);
	done_1999 = true;
}

void do_entry_command(char *buf)
{
	switch (buf[0]) {
	case 'p':
		do_prototype(&buf[2]);
		break;
	case 'd':
		do_description(buf);
		break;
	case 'm':
		do_implementation(buf);
		break;
	case 'e':
		do_example(buf);
		break;
	case 'r':
		do_iso1990_ref(&buf[2]);
		break;
	case 'R':
		do_iso1999_ref(&buf[2]);
		break;
	case 'L':
		break;
	case 'H':
		fprintf(out_file, "%s\n", &buf[2]);
		break;
	default:
		fprintf(stderr, "%d: Invalid command `%c'\n", lineno, buf[0]);
	}
}

void do_entry(char *buf)
{
	char filename[24];
	static int entry_num = 0;
	static char last_entry[48];

	fprintf(sym_file, "%s\n", buf);
	sprintf(filename, "%s.html", buf);
	out_file = fopen(filename, "w");
	fprintf(out_file, HTML_HEADER);

	done_1990 = done_1999 = false;

	strcpy(filename, header_name);
	*strchr(filename, '.') = '\0';
	fprintf(out_file,
		"<!-- Buttons -->\n"
		"<table cellspacing=\"0\" cellpadding=\"0\">\n"
		"<tr>\n"
		"<td><a href=\"../index.html\"><img src=\"../images/home.gif\" alt=\"Return to index\" border=\"0\" height=\"32\" width=\"32\"></a></td>\n"
		"<td><a href=\"../%s.html\"><img src=\"../images/up.gif\" alt=\"Return to header\" border=\"0\" height=\"32\" width=\"32\"></a></td>\n",
		filename);

	if (entry_num++ > 0)
		fprintf(out_file,
			"<td><a href=\"%s.html\"><img src=\"../images/left.gif\" alt=\"Previous symbol\" border=\"0\" height=\"32\" width=\"32\"></a></td>\n", last_entry);
	else
		fprintf(out_file, 
			"<td><img src=\"../images/left-off.gif\" alt=\"Previous symbol\" height=\"32\" width=\"32\"></td>\n");
	fprintf(out_file,
		"<td>@next entry@</td>\n");

	fprintf(out_file,
		"</tr>\n"
		"</table>\n"
		"<!-- End Buttons -->\n");
	fprintf(out_file, "<h1>%s</h1>\n", buf);
	strcpy(last_entry, buf);

	while (getline(buf) != NULL)
		if (buf[0] == '@') {
			if (buf[1] == ':' || buf[1] == 'x') {
				ungetline(buf);
				break;
			} else
				do_entry_command(&buf[1]);
		} else if (buf[0] != '\0')
			fprintf(stderr, "%d: Invalid line `%s'\n", lineno, buf);

	fprintf(out_file,
		"\n</body>\n"
		"</html>\n");
	fclose(out_file);
}

void do_command(char *buf)
{
	switch (buf[0]) {
	case 'h':
		strcpy(header_name, &buf[2]);
		break;
	case 'x':
		do_xref(&buf[2]);
		break;
	case ':':
		do_entry(&buf[2]);
		break;
	case 'L':
		break;
	case 'H':
/*		fprintf(out_file, "%s\n", &buf[2]); */
		break;
	default:
		fprintf(stderr, "%d: Invalid command `%c'\n", lineno, buf[0]);
	}
}

int main(void)
{
	char buf[MAXLINELEN];

	sym_file = fopen("symbols.idx", "w");

	while (getline(buf) != NULL)
		if (buf[0] == '@')
			do_command(&buf[1]);

	fclose(sym_file);

	return 0;
}
